/*
 * bsc1176931_eth
 *
 * Fix for CVE-2020-0429, bsc#1176931 (l2tp_eth.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if !IS_MODULE(CONFIG_L2TP_ETH)
#error "Live patch supports only CONFIG_L2TP_ETH=m"
#endif

#define KGR_PATCHED_MODULE "l2tp_eth"

/* klp-ccp: from net/l2tp/l2tp_eth.c */
#define pr_fmt(fmt) KGR_PATCHED_MODULE ": " fmt

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/l2tp.h>
#include <linux/in.h>
#include <linux/etherdevice.h>
#include <linux/spinlock.h>
#include <net/tcp_states.h>
#include <net/net_namespace.h>
#include <net/netns/generic.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1176931.h"
#include "../kallsyms_relocs.h"

#include "bsc1176931_common.h"

static struct module *kgrp_l2tp_eth_mod;

/* klp-ccp: from net/l2tp/l2tp_core.h */
static struct l2tp_session *(*kgre_l2tp_session_find)(struct net *net,
				       struct l2tp_tunnel *tunnel,
				       u32 session_id);

static struct l2tp_tunnel *(*kgre_l2tp_tunnel_find)(struct net *net, u32 tunnel_id);

static int (*kgre_l2tp_session_delete)(struct l2tp_session *session);

/* klp-ccp: from net/l2tp/l2tp_eth.c */
#define L2TP_ETH_DEV_NAME	"l2tpeth%d"

struct l2tp_eth {
	struct net_device	*dev;
	struct sock		*tunnel_sock;
	struct l2tp_session	*session;
	struct list_head	list;
	atomic_long_t		tx_bytes;
	atomic_long_t		tx_packets;
	atomic_long_t		tx_dropped;
	atomic_long_t		rx_bytes;
	atomic_long_t		rx_packets;
	atomic_long_t		rx_errors;
};

struct l2tp_eth_sess {
	struct net_device	*dev;
};

static unsigned int (*kgre_l2tp_eth_net_id);
struct l2tp_eth_net {
	struct list_head l2tp_eth_dev_list;
	spinlock_t l2tp_eth_lock;
};

static inline struct l2tp_eth_net *kgrr_l2tp_eth_pernet(struct net *net)
{
	return net_generic(net, (*kgre_l2tp_eth_net_id));
}

static void (*kgre_l2tp_eth_dev_setup)(struct net_device *dev);

static void (*kgre_l2tp_eth_dev_recv)(struct l2tp_session *session, struct sk_buff *skb, int data_len);

static void (*kgre_l2tp_eth_delete)(struct l2tp_session *session);

#if defined(CONFIG_L2TP_DEBUGFS) || defined(CONFIG_L2TP_DEBUGFS_MODULE)
static void (*kgre_l2tp_eth_show)(struct seq_file *m, void *arg);

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

int kgrp_l2tp_eth_create(struct net *net, u32 tunnel_id, u32 session_id, u32 peer_session_id, struct l2tp_session_cfg *cfg)
{
	struct net_device *dev;
	char name[IFNAMSIZ];
	struct l2tp_tunnel *tunnel;
	struct l2tp_session *session;
	struct l2tp_eth *priv;
	struct l2tp_eth_sess *spriv;
	int rc;
	struct l2tp_eth_net *pn;

	tunnel = (*kgre_l2tp_tunnel_find)(net, tunnel_id);
	if (!tunnel) {
		rc = -ENODEV;
		goto out;
	}

	/*
	 * Fix CVE-2020-0429
	 *  -5 lines
	 */

	if (cfg->ifname) {
		dev = dev_get_by_name(net, cfg->ifname);
		if (dev) {
			dev_put(dev);
			rc = -EEXIST;
			goto out;
		}
		strlcpy(name, cfg->ifname, IFNAMSIZ);
	} else
		strcpy(name, L2TP_ETH_DEV_NAME);

	session = kgrp_l2tp_session_create(sizeof(*spriv), tunnel, session_id,
				      peer_session_id, cfg);
	/*
	 * Fix CVE-2020-0429
	 *  -2 lines, +2 lines
	 */
	if (IS_ERR(session)) {
		rc = PTR_ERR(session);
		goto out;
	}

	dev = alloc_netdev(sizeof(*priv), name, NET_NAME_UNKNOWN,
			   (*kgre_l2tp_eth_dev_setup));
	if (!dev) {
		rc = -ENOMEM;
		goto out_del_session;
	}

	dev_net_set(dev, net);
	if (session->mtu == 0)
		session->mtu = dev->mtu - session->hdr_len;
	dev->mtu = session->mtu;
	dev->needed_headroom += session->hdr_len;

	priv = netdev_priv(dev);
	priv->dev = dev;
	priv->session = session;
	INIT_LIST_HEAD(&priv->list);

	priv->tunnel_sock = tunnel->sock;
	session->recv_skb = (*kgre_l2tp_eth_dev_recv);
	session->session_close = (*kgre_l2tp_eth_delete);
#if defined(CONFIG_L2TP_DEBUGFS) || defined(CONFIG_L2TP_DEBUGFS_MODULE)
	session->show = (*kgre_l2tp_eth_show);
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
	spriv = l2tp_session_priv(session);
	spriv->dev = dev;

	rc = register_netdev(dev);
	if (rc < 0)
		goto out_del_dev;

	/*
	 * Fix CVE-2020-25645
	 *  -1 line, +1 line
	 * Necessary adaption for livepatch.
	 */
	__module_get(kgrp_l2tp_eth_mod);
	/* Must be done after register_netdev() */
	strlcpy(session->ifname, dev->name, IFNAMSIZ);

	dev_hold(dev);
	pn = kgrr_l2tp_eth_pernet(dev_net(dev));
	spin_lock(&pn->l2tp_eth_lock);
	list_add(&priv->list, &pn->l2tp_eth_dev_list);
	spin_unlock(&pn->l2tp_eth_lock);

	return 0;

out_del_dev:
	free_netdev(dev);
	spriv->dev = NULL;
out_del_session:
	(*kgre_l2tp_session_delete)(session);
out:
	return rc;
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "l2tp_eth_net_id", (void *)&kgre_l2tp_eth_net_id, "l2tp_eth" },
	{ "l2tp_session_find", (void *)&kgre_l2tp_session_find, "l2tp_core" },
	{ "l2tp_tunnel_find", (void *)&kgre_l2tp_tunnel_find, "l2tp_core" },
	{ "l2tp_session_delete", (void *)&kgre_l2tp_session_delete,
	  "l2tp_core" },
	{ "l2tp_eth_dev_setup", (void *)&kgre_l2tp_eth_dev_setup, "l2tp_eth" },
	{ "l2tp_eth_dev_recv", (void *)&kgre_l2tp_eth_dev_recv, "l2tp_eth" },
	{ "l2tp_eth_delete", (void *)&kgre_l2tp_eth_delete, "l2tp_eth" },
	{ "l2tp_eth_show", (void *)&kgre_l2tp_eth_show, "l2tp_eth" },
};

static int kgr_patch_bsc1176931_eth_module_notify(struct notifier_block *nb,
						  unsigned long action,
						void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	kgrp_l2tp_eth_mod = mod;
	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1176931_eth_module_nb = {
	.notifier_call = kgr_patch_bsc1176931_eth_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1176931_eth_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	kgrp_l2tp_eth_mod = find_module(KGR_PATCHED_MODULE);
	if (kgrp_l2tp_eth_mod) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1176931_eth_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1176931_eth_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1176931_eth_module_nb);
}
